@@ -18,7 +18,6 @@
 */
 
 #include "Models/StateSpace/Filters/SparseMatrix.hpp"
-#include <iostream>
 #include <utility>
 #include "LinAlg/SpdMatrix.hpp"
 #include "LinAlg/DiagonalMatrix.hpp"
@@ -43,6 +42,369 @@ namespace BOOM {
     report_error(err.str());
   }
 
+  void SparseKalmanMatrix::check_can_multiply(int vector_size) const {
+    conforms_to_cols(vector_size);
+  }
+
+  void SparseKalmanMatrix::check_can_Tmult(int vector_size) const {
+    conforms_to_rows(vector_size);
+  }
+
+  void SparseKalmanMatrix::check_can_add(int rows, int cols) const {
+    conforms_to_rows(rows);
+    conforms_to_cols(cols);
+  }
+
+  SparseMatrixProduct::SparseMatrixProduct() {}
+
+  void SparseMatrixProduct::add_term(const Ptr<SparseKalmanMatrix> &term,
+                                     bool transpose) {
+    check_term(term, transpose);
+    terms_.push_back(term);
+    transposed_.push_back(transpose);
+  }
+
+  int SparseMatrixProduct::nrow() const {
+    if (terms_.empty()) {
+      return 0;
+    }
+    return transposed_[0] ? terms_[0]->ncol() : terms_[0]->nrow();
+  }
+
+  int SparseMatrixProduct::ncol() const {
+    if (terms_.empty()) {
+      return 0;
+    }
+    return transposed_.back() ? terms_.back()->nrow() : terms_.back()->ncol();
+  }
+
+  Vector SparseMatrixProduct::operator*(const ConstVectorView &v) const {
+    Vector ans = v;
+    for (int i = terms_.size() - 1; i >= 0; --i) {
+      if (transposed_[i]) {
+        ans = terms_[i]->Tmult(ans);
+      } else {
+        ans = (*terms_[i]) * ans;
+      }
+    }
+    return ans;
+  }
+
+  Vector SparseMatrixProduct::operator*(const Vector &rhs) const {
+    return (*this) * ConstVectorView(rhs);
+  }
+
+  Vector SparseMatrixProduct::operator*(const VectorView &rhs) const {
+    return (*this) * ConstVectorView(rhs);
+  }
+
+  Matrix SparseMatrixProduct::operator*(const Matrix &rhs) const {
+    Matrix ans(rhs);
+    for (int i = terms_.size() - 1; i >= 0; --i) {
+      if (transposed_[i]) {
+        ans = terms_[i]->Tmult(ans);
+      } else {
+        ans = (*terms_[i]) * ans;
+      }
+    }
+    return ans;
+  }
+
+  Vector SparseMatrixProduct::Tmult(const ConstVectorView &rhs) const {
+    Vector ans(rhs);
+    for (size_t i = 0; i < terms_.size(); ++i) {
+      if (transposed_[i]) {
+        ans = *terms_[i] * ans;
+      } else {
+        ans = terms_[i]->Tmult(ans);
+      }
+    }
+    return ans;
+  }
+
+  Matrix SparseMatrixProduct::Tmult(const Matrix &rhs) const {
+    Matrix ans(rhs);
+    for (size_t i = 0; i < terms_.size(); ++i) {
+      if (transposed_[i]) {
+        ans = *terms_[i] * ans;
+      } else {
+        ans = terms_[i]->Tmult(ans);
+      }
+    }
+    return ans;
+  }
+
+  Ptr<SparseMatrixProduct>
+  SparseMatrixProduct::sparse_sandwich(const SpdMatrix &N) const {
+    NEW(DenseSpd, SparseN)(N);
+    NEW(SparseMatrixProduct, ans)();
+    for (int i = terms_.size() - 1; i >= 0; --i) {
+      ans->add_term(terms_[i], !transposed_[i]);
+    }
+    ans->add_term(SparseN);
+    for (int i = 0; i < terms_.size(); ++i) {
+      ans->add_term(terms_[i], transposed_[i]);
+    }
+    return ans;
+  }
+
+  Ptr<SparseMatrixProduct>
+  SparseMatrixProduct::sparse_sandwich_transpose(const SpdMatrix &N) const {
+    NEW(DenseSpd, SparseN)(N);
+    NEW(SparseMatrixProduct, ans)();
+    for (int i = 0; i < terms_.size(); ++i) {
+      ans->add_term(terms_[i], transposed_[i]);
+    }
+    ans->add_term(SparseN);
+    for(int i = terms_.size() - 1; i >= 0; --i) {
+      ans->add_term(terms_[i], !transposed_[i]);
+    }
+    return ans;
+  }
+
+  Vector SparseMatrixProduct::diag() const {
+    int m = std::min(nrow(), ncol());
+    Vector ans(m);
+    for (int i = 0; i < m; ++i) {
+      Vector one(ncol(), 0.0);
+      one[i] = 1.0;
+      ans[i] = ((*this) * one)[i];
+    }
+    return ans;
+  }
+
+  SpdMatrix SparseMatrixProduct::inner() const {
+    SpdMatrix ans(nrow(), 1.0);
+    for (int i = 0; i < terms_.size(); ++i) {
+      if (transposed_[i]) {
+        ans = terms_[i]->sandwich(ans);
+      } else {
+        ans = terms_[i]->sandwich_transpose(ans);
+      }
+    }
+    return ans;
+  }
+
+  SpdMatrix SparseMatrixProduct::inner(const ConstVectorView &weights) const {
+    SpdMatrix ans(weights.size());
+    ans.diag() = weights;
+
+    for (int i = 0; i < terms_.size(); ++i) {
+      if (transposed_[i]) {
+        ans = terms_[i]->sandwich(ans);
+      } else {
+        ans = terms_[i]->sandwich_transpose(ans);
+      }
+    }
+    return ans;
+  }
+
+  Matrix &SparseMatrixProduct::add_to(Matrix &P) const {
+    P += this->dense();
+    return P;
+  }
+
+  Matrix SparseMatrixProduct::dense() const {
+    SpdMatrix Id(ncol(), 1.0);
+    return (*this) * Id;
+  }
+
+  void SparseMatrixProduct::check_term(const Ptr<SparseKalmanMatrix> &term,
+                                       bool transpose) {
+    if (terms_.empty()) {
+      return;
+    }
+    bool final_transpose = transposed_.back();
+    const Ptr<SparseKalmanMatrix>& final_term(terms_.back());
+    size_t final_dim =
+        final_transpose ? final_term->nrow() : final_term->ncol();
+    size_t leading_dim = transpose ? term->ncol() : term-> nrow();
+    if (final_dim != leading_dim) {
+      std::ostringstream err;
+      err << "Incompatible matrix following term " << terms_.size()
+          << ".  Final dimension of previous term: "
+          << final_dim
+          << ".  Leading dimension of new term: " << leading_dim
+          << ".";
+      report_error(err.str());
+    }
+  }
+
+  //===========================================================================
+
+  SparseMatrixSum::SparseMatrixSum() {}
+
+  void SparseMatrixSum::add_term(
+      const Ptr<SparseKalmanMatrix> &term, double coefficient) {
+    if (!terms_.empty()) {
+      if (term->nrow() != terms_.back()->nrow()
+          || term->ncol() != terms_.back()->ncol()) {
+        report_error("Incompatible sparse matrices in sum.");
+      }
+    }
+    terms_.push_back(term);
+    coefficients_.push_back(coefficient);
+  }
+
+  int SparseMatrixSum::nrow() const {
+    if (terms_.empty()) {
+      return 0;
+    }
+    return terms_.back()->nrow();
+  }
+
+  int SparseMatrixSum::ncol() const {
+    if (terms_.empty()) {
+      return 0;
+    }
+    return terms_.back()->ncol();
+  }
+
+  Vector SparseMatrixSum::operator*(const ConstVectorView &rhs) const {
+    Vector ans(nrow(), 0.0);
+    for (int i = 0; i < terms_.size(); ++i) {
+      ans += coefficients_[i] * ((*terms_[i]) * rhs);
+    }
+    return ans;
+  }
+
+  Vector SparseMatrixSum::operator*(const Vector &rhs) const {
+    return (*this) * ConstVectorView(rhs);
+  }
+
+  Vector SparseMatrixSum::operator*(const VectorView &rhs) const {
+    return (*this) * ConstVectorView(rhs);
+  }
+
+  Matrix SparseMatrixSum::operator*(const Matrix &rhs) const {
+    Matrix ans(nrow(), rhs.ncol(), 0.0);
+    for (int i = 0; i < terms_.size(); ++i) {
+      ans += coefficients_[i] * ((*terms_[i]) * rhs);
+    }
+    return ans;
+  }
+
+  Vector SparseMatrixSum::Tmult(const ConstVectorView &rhs) const {
+    Vector ans(ncol());
+    for (int i = 0; i < terms_.size(); ++i) {
+      ans += coefficients_[i] * terms_[i]->Tmult(rhs);
+    }
+    return ans;
+  }
+
+  Matrix SparseMatrixSum::Tmult(const Matrix &rhs) const {
+    Matrix ans(ncol(), rhs.ncol());
+    for (int i = 0; i < terms_.size(); ++i) {
+      ans += coefficients_[i] * terms_[i]->Tmult(rhs);
+    }
+    return ans;
+  }
+
+  SpdMatrix SparseMatrixSum::inner() const {
+    return dense().inner();
+  }
+
+  SpdMatrix SparseMatrixSum::inner(const ConstVectorView &weights) const {
+    return dense().inner(weights);
+  }
+
+  Matrix &SparseMatrixSum::add_to(Matrix &rhs) const {
+    for (int i = 0; i < terms_.size(); ++i) {
+      if (coefficients_[i] != 0.0) {
+        rhs /= coefficients_[i];
+        terms_[i]->add_to(rhs);
+        rhs *= coefficients_[i];
+      }
+    }
+    return rhs;
+  }
+
+  //===========================================================================
+
+  SparseBinomialInverse::SparseBinomialInverse(
+      const Ptr<SparseKalmanMatrix> &Ainv,
+      const Ptr<SparseKalmanMatrix> &U,
+      const SpdMatrix &B,
+      double Ainv_logdet)
+      : Ainv_(Ainv),
+        U_(U),
+        B_(B)
+  {
+    SparseMatrixProduct tmp;
+    tmp.add_term(U, true);
+    tmp.add_term(Ainv);
+    tmp.add_term(U);
+
+    inner_matrix_ = SpdMatrix(B.nrow(), 1.0);
+    inner_matrix_ += B * tmp.dense();
+    inner_matrix_ = inner_matrix_.inv();
+
+    logdet_ = Ainv_logdet + inner_matrix_.logdet();
+  }
+
+  SparseBinomialInverse::SparseBinomialInverse(
+      const Ptr<SparseKalmanMatrix> &Ainv,
+      const Ptr<SparseKalmanMatrix> &U,
+      const SpdMatrix &B,
+      const Matrix &inner,
+      double logdet)
+      : Ainv_(Ainv),
+        U_(U),
+        B_(B),
+        inner_matrix_(inner),
+        logdet_(logdet)
+  {}
+
+  Vector SparseBinomialInverse::operator*(const ConstVectorView &rhs) const {
+    Vector ans = (*Ainv_) * rhs;
+    ans -= (*Ainv_) * (*U_ * (inner_matrix_ * (B_ * (U_->Tmult(*Ainv_ * rhs)))));
+    return ans;
+  }
+
+  Vector SparseBinomialInverse::operator*(const Vector &rhs) const {
+    return (*this) * ConstVectorView(rhs);
+  }
+
+  Vector SparseBinomialInverse::operator*(const VectorView &rhs) const {
+    return (*this) * ConstVectorView(rhs);
+  }
+
+  Matrix SparseBinomialInverse::operator*(const Matrix &rhs) const {
+    Matrix ans = *Ainv_ * rhs;
+    ans -= *Ainv_ * (*U_ * (inner_matrix_ * (B_ * (U_->Tmult(*Ainv_ * rhs)))));
+    return ans;
+  }
+
+  Vector SparseBinomialInverse::Tmult(const ConstVectorView &rhs) const {
+    return (*this) * rhs;
+  }
+
+  Matrix SparseBinomialInverse::Tmult(const Matrix &rhs) const {
+    return (*this) * rhs;
+  }
+
+  Matrix SparseBinomialInverse::dense() const {
+    SpdMatrix I(ncol(), 1.0);
+    return (*this) * I;
+  }
+
+  Matrix & SparseBinomialInverse::add_to(Matrix &rhs) const {
+    rhs += this->dense();
+    return rhs;
+  }
+
+  SpdMatrix SparseBinomialInverse::inner() const {
+    return this->dense().inner();
+  }
+  SpdMatrix SparseBinomialInverse::inner(const ConstVectorView &weights) const {
+    return this->dense().inner(weights);
+  }
+
+  double SparseBinomialInverse::logdet() const {
+    return logdet_;
+  }
+  //===========================================================================
+
   namespace {
     template <class VECTOR>
     Vector sparse_multiply_impl(const SparseMatrixBlock &m, const VECTOR &v) {
@@ -53,6 +415,16 @@ namespace BOOM {
     }
   }  // namespace
 
+  void SparseMatrixBlock::check_can_multiply(
+      const VectorView &lhs, const ConstVectorView &rhs) const {
+    if (lhs.size() != nrow()) {
+      report_error("Left hand side is the wrong dimension.");
+    }
+    if (rhs.size() != ncol()) {
+      report_error("Right hand side is the wrong dimension.");
+    }
+  }
+
   Vector SparseMatrixBlock::operator*(const Vector &v) const {
     return sparse_multiply_impl(*this, v);
   }
@@ -178,16 +550,6 @@ namespace BOOM {
     blocks_.push_back(block);
   }
 
-  void BlockDiagonalMatrixBlock::check_can_multiply(
-      const VectorView &lhs, const ConstVectorView &rhs) const {
-    if (lhs.size() != dim_) {
-      report_error("Left hand side is the wrong dimension.");
-    }
-    if (rhs.size() != dim_) {
-      report_error("Right hand side is the wrong dimension.");
-    }
-  }
-
   void BlockDiagonalMatrixBlock::multiply(VectorView lhs,
                                           const ConstVectorView &rhs) const {
     check_can_multiply(lhs, rhs);
@@ -1106,6 +1468,11 @@ namespace BOOM {
   }
 
   //======================================================================
+
+  StackedRegressionCoefficients *StackedRegressionCoefficients::clone() const {
+    return new StackedRegressionCoefficients(*this);
+  }
+
   void StackedRegressionCoefficients::add_row(const Ptr<GlmCoefs> &beta) {
     if (!coefficients_.empty()) {
       if (beta->nvars_possible() != coefficients_[0]->nvars_possible()) {
@@ -1127,6 +1494,56 @@ namespace BOOM {
     }
   }  // namespace
 
+  void StackedRegressionCoefficients::multiply(
+      VectorView lhs, const ConstVectorView &rhs) const {
+    check_can_multiply(lhs, rhs);
+    for (int i = 0; i < lhs.size(); ++i) {
+      lhs[i] = coefficients_[i]->predict(rhs);
+    }
+  }
+
+  void StackedRegressionCoefficients::multiply_and_add(
+      VectorView lhs, const ConstVectorView &rhs) const {
+    check_can_multiply(rhs.size());
+    if (lhs.size() != nrow()) {
+      report_error("lhs argument is the wrong size in "
+                   "StackedRegressionCoefficients::multiply_and_add.");
+    }
+    for (int i = 0; i < lhs.size(); ++i) {
+      lhs[i] += coefficients_[i]->predict(rhs);
+    }
+  }
+
+  void StackedRegressionCoefficients::add_to_block(SubMatrix block) const {
+    for (int i = 0; i < block.nrow(); ++i) {
+      coefficients_[i]->add_to(block.row(i));
+    }
+  }
+
+  void StackedRegressionCoefficients::Tmult(
+      VectorView lhs, const ConstVectorView &rhs) const {
+    check_can_Tmult(rhs.size());
+    if (lhs.size() != ncol()) {
+      report_error("lhs argument is the wrong size in "
+                   "StackedRegressionCoefficients::Tmult.");
+    }
+    for (size_t i = 0; i < lhs.size(); ++i) {
+      lhs[i] = 0;
+      for (size_t j = 0; j < rhs.size(); ++j) {
+        lhs[i] += coefficients_[j]->value()[i] * rhs[j];
+      }
+    }
+  }
+
+  void StackedRegressionCoefficients::multiply_inplace(
+      VectorView x) const {
+    check_can_multiply(x.size());
+    if (nrow() != ncol()) {
+      report_error("multiply_inplace only applies to square matrices.");
+    }
+    x = *this * x;
+  }
+
   Vector StackedRegressionCoefficients::operator*(
       const Vector &v) const {
     return stacked_regression_vector_mult(v, *this);
@@ -1143,12 +1560,7 @@ namespace BOOM {
   Vector StackedRegressionCoefficients::Tmult(
       const ConstVectorView &x) const {
     Vector ans(ncol());
-    for (int i = 0; i < ncol(); ++i) {
-      ans[i] = 0;
-      for (int j = 0; j < nrow(); ++j) {
-        ans[i] += coefficients_[j]->value()[i] * x[i];
-      }
-    }
+    this->Tmult(VectorView(ans), x);
     return ans;
   }
 
@@ -1314,6 +1726,29 @@ namespace BOOM {
   //======================================================================
   BlockDiagonalMatrix::BlockDiagonalMatrix() : nrow_(0), ncol_(0) {}
 
+  BlockDiagonalMatrix::BlockDiagonalMatrix(const BlockDiagonalMatrix &rhs)
+      : nrow_(0),
+        ncol_(0)
+  {
+    for (const auto &block : rhs.blocks_) {
+      add_block(block->clone());
+    }
+  }
+
+  BlockDiagonalMatrix & BlockDiagonalMatrix::operator=(const BlockDiagonalMatrix &rhs) {
+    if (&rhs != this) {
+      clear();
+      for (const auto &block : rhs.blocks_) {
+        add_block(block->clone());
+      }
+    }
+    return *this;
+  }
+
+  BlockDiagonalMatrix * BlockDiagonalMatrix::clone() const {
+    return new BlockDiagonalMatrix(*this);
+  }
+
   void BlockDiagonalMatrix::add_block(const Ptr<SparseMatrixBlock> &m) {
     blocks_.push_back(m);
     nrow_ += m->nrow();
@@ -1344,35 +1779,63 @@ namespace BOOM {
   //---------------------------------------------------------------------------
   // TODO(steve): add a unit test for the case where diagonal
   // blocks are not square.
-  Vector block_multiply(const ConstVectorView &v, int nrow, int ncol,
-                        const std::vector<Ptr<SparseMatrixBlock>> &blocks_) {
+  void block_multiply_view(VectorView ans, const ConstVectorView &v, int nrow, int ncol,
+                           const std::vector<Ptr<SparseMatrixBlock>> &blocks) {
     if (v.size() != ncol) {
-      report_error(
-          "incompatible vector in "
-          "BlockDiagonalMatrix::operator*");
+      report_error("incompatible vector in BlockDiagonalMatrix::operator*");
     }
-    Vector ans(nrow);
-
     int lhs_pos = 0;
     int rhs_pos = 0;
-    for (int b = 0; b < blocks_.size(); ++b) {
-      int nr = blocks_[b]->nrow();
+    for (int b = 0; b < blocks.size(); ++b) {
+      int nr = blocks[b]->nrow();
       VectorView lhs(ans, lhs_pos, nr);
       lhs_pos += nr;
 
-      int nc = blocks_[b]->ncol();
+      int nc = blocks[b]->ncol();
       if (nc > 0) {
         ConstVectorView rhs(v, rhs_pos, nc);
         rhs_pos += nc;
-        blocks_[b]->multiply(lhs, rhs);
+        blocks[b]->multiply(lhs, rhs);
       } else {
         lhs = 0.0;
       }
     }
+  }
+
+  Vector block_multiply(const ConstVectorView &v, int nrow, int ncol,
+                        const std::vector<Ptr<SparseMatrixBlock>> &blocks) {
+    Vector ans(nrow);
+    block_multiply_view(VectorView(ans), v, nrow, ncol, blocks);
     return ans;
   }
 
-  Vector block_transpose_multiply(
+  void block_multiply_and_add(
+      VectorView ans,
+      const ConstVectorView &v,
+      int nrow,
+      int ncol,
+      const std::vector<Ptr<SparseMatrixBlock>> &blocks) {
+    if (v.size() != ncol) {
+      report_error("incompatible vector in BlockDiagonalMatrix::operator*");
+    }
+    int lhs_pos = 0;
+    int rhs_pos = 0;
+    for (int b = 0; b < blocks.size(); ++b) {
+      int nr = blocks[b]->nrow();
+      VectorView lhs(ans, lhs_pos, nr);
+      lhs_pos += nr;
+
+      int nc = blocks[b]->ncol();
+      if (nc > 0) {
+        ConstVectorView rhs(v, rhs_pos, nc);
+        rhs_pos += nc;
+        blocks[b]->multiply_and_add(lhs, rhs);
+      }
+    }
+  }
+
+  void block_transpose_multiply_view(
+      VectorView lhs,
       const ConstVectorView &v,
       int nrow,
       int ncol,
@@ -1381,24 +1844,62 @@ namespace BOOM {
     if (v.size() != nrow) {
       report_error("incompatible vector in Tmult");
     }
+    if (lhs.size() != ncol) {
+      report_error("Incompatible LHS in block_transpose_multiply.");
+    }
 
     int lhs_pos = 0;
     int rhs_pos = 0;
-    Vector ans(ncol, 0);
 
     for (int b = 0; b < blocks.size(); ++b) {
-      VectorView lhs(ans, lhs_pos, blocks[b]->ncol());
+      VectorView lhs_chunk(lhs, lhs_pos, blocks[b]->ncol());
       lhs_pos += blocks[b]->ncol();
-      ConstVectorView rhs(v, rhs_pos, blocks[b]->nrow());
+      ConstVectorView rhs_chunk(v, rhs_pos, blocks[b]->nrow());
       rhs_pos += blocks[b]->nrow();
-      blocks[b]->Tmult(lhs, rhs);
+      blocks[b]->Tmult(lhs_chunk, rhs_chunk);
     }
-    return ans;
-
   }
 
+  Vector block_transpose_multiply(
+      const ConstVectorView &v,
+      int nrow,
+      int ncol,
+      const std::vector<Ptr<SparseMatrixBlock>> &blocks) {
+    Vector ans(ncol, 0);
+    block_transpose_multiply_view(VectorView(ans), v, nrow, ncol, blocks);
+    return ans;
+  }
+
+  void block_multiply_inplace(VectorView x, int nrow, int ncol,
+                              const std::vector<Ptr<SparseMatrixBlock>> & blocks) {
+    if (nrow != ncol) {
+      report_error("multiply_inplace only works for square matrices.");
+    }
+    int start = 0;
+    for (const auto &block : blocks) {
+      if (block->nrow() != block->ncol()) {
+        report_error("All individual blocks must be square for multiply_inplace.");
+      }
+      VectorView chunk(x, start, block->ncol());
+      block->multiply_inplace(chunk);
+      start += block->nrow();
+    }
+  }
 
   //---------------------------------------------------------------------------
+  void BlockDiagonalMatrix::multiply(VectorView lhs, const ConstVectorView &rhs) const {
+    block_multiply_view(lhs, rhs, nrow(), ncol(), blocks_);
+  }
+
+  void BlockDiagonalMatrix::multiply_and_add(VectorView lhs,
+                                             const ConstVectorView &rhs) const {
+    block_multiply_and_add(lhs, rhs, nrow(), ncol(), blocks_);
+  }
+
+  void BlockDiagonalMatrix::multiply_inplace(VectorView v) const {
+    block_multiply_inplace(v, nrow(), ncol(), blocks_);
+  }
+
   Vector BlockDiagonalMatrix::operator*(const Vector &v) const {
     return block_multiply(ConstVectorView(v), nrow(), ncol(), blocks_);
   }
@@ -1410,6 +1911,10 @@ namespace BOOM {
     return block_multiply(v, nrow(), ncol(), blocks_);
   }
 
+  void BlockDiagonalMatrix::Tmult(VectorView lhs, const ConstVectorView &rhs) const {
+    block_transpose_multiply_view(lhs, rhs, nrow(), ncol(), blocks_);
+  }
+
   Vector BlockDiagonalMatrix::Tmult(const ConstVectorView &x) const {
     return block_transpose_multiply(x, nrow(), ncol(), blocks_);
   }
@@ -1730,26 +2235,6 @@ namespace BOOM {
     return P;
   }
 
-  void SparseVerticalStripMatrix::check_can_multiply(int vector_size) const {
-    if (ncol() != vector_size) {
-      report_error("Incompatible vector multiplication.");
-    }
-  }
-  void SparseVerticalStripMatrix::check_can_Tmult(int vector_size) const {
-    if (nrow() != vector_size) {
-      std::ostringstream err;
-      err << "Incompatible vector (transpose-)multiplication.  "
-          << "This matrix has " << nrow() << " rows.  The target vector has "
-          << vector_size << " elements." << std::endl;
-      report_error(err.str());
-    }
-  }
-
-  void SparseVerticalStripMatrix::check_can_add(int rows, int cols) const {
-    if (nrow() != rows || ncol() != cols) {
-      report_error("Incompatible matrix addition.");
-    }
-  }
 
   //===========================================================================
   // LHS = *this * RHS
@@ -1759,6 +2244,30 @@ namespace BOOM {
         ncol_(0)
   {}
 
+  ErrorExpanderMatrix::ErrorExpanderMatrix(const ErrorExpanderMatrix &rhs)
+      : nrow_(0),
+        ncol_(0)
+  {
+    for (const auto &b : rhs.blocks_) {
+      add_block(b->clone());
+    }
+  }
+
+  ErrorExpanderMatrix & ErrorExpanderMatrix::operator=(
+      const ErrorExpanderMatrix &rhs) {
+    if (&rhs != this) {
+      clear();
+      for (const auto &b : rhs.blocks_) {
+        add_block(b->clone());
+      }
+    }
+    return *this;
+  }
+
+  ErrorExpanderMatrix * ErrorExpanderMatrix::clone() const {
+    return new ErrorExpanderMatrix(*this);
+  }
+
   int ErrorExpanderMatrix::nrow() const {
     return nrow_;
   }
@@ -1772,6 +2281,12 @@ namespace BOOM {
     increment_sizes(block);
   }
 
+  // void ErrorExpanderMatrix::add_block(const Ptr<ErrorExpanderMatrix> &blocks) {
+  //   for (const auto &block : blocks->blocks_) {
+  //     add_block(block);
+  //   }
+  // }
+
   void ErrorExpanderMatrix::replace_block(
       int block_index,
       const Ptr<SparseMatrixBlock> &block) {
@@ -1809,6 +2324,45 @@ namespace BOOM {
     recompute_sizes();
   }
 
+  // This logic is almost the same as the 'block_multiply' function above.
+  void ErrorExpanderMatrix::multiply(VectorView lhs,
+                                     const ConstVectorView &rhs) const {
+    int lhs_pos = 0;
+    int rhs_pos = 0;
+    for (const auto &block : blocks_) {
+      int nr = block->nrow();
+      VectorView left_block(lhs, lhs_pos, nr);
+
+      int nc = block->ncol();
+      if (nc > 0) {
+        const ConstVectorView right_block(rhs, rhs_pos, nc);
+        block->multiply(left_block, right_block);
+      } else {
+        left_block = 0;
+      }
+      lhs_pos += nr;
+      rhs_pos += nc;
+    }
+  }
+
+  void ErrorExpanderMatrix::multiply_and_add(
+      VectorView lhs, const ConstVectorView &rhs) const {
+    int lhs_pos = 0;
+    int rhs_pos = 0;
+    for (const auto &block : blocks_) {
+      int nr = block->nrow();
+      VectorView left_block(lhs, lhs_pos, nr);
+
+      int nc = block->ncol();
+      if (nc > 0) {
+        const ConstVectorView right_block(rhs, rhs_pos, nc);
+        block->multiply_and_add(left_block, right_block);
+      }
+      lhs_pos += nr;
+      rhs_pos += nc;
+    }
+  }
+
   Vector ErrorExpanderMatrix::operator*(const Vector &v) const {
     return block_multiply(v, nrow(), ncol(), blocks_);
   }
@@ -1825,6 +2379,36 @@ namespace BOOM {
     return block_transpose_multiply(x, nrow(), ncol(), blocks_);
   }
 
+  void ErrorExpanderMatrix::Tmult(VectorView lhs,
+                                  const ConstVectorView &rhs) const {
+    block_transpose_multiply_view(lhs, rhs, nrow(), ncol(), blocks_);
+  }
+
+  void ErrorExpanderMatrix::multiply_inplace(VectorView x) const {
+    block_multiply_inplace(x, nrow(), ncol(), blocks_);
+  }
+
+  void ErrorExpanderMatrix::add_to_block(SubMatrix block) const {
+    if (block.nrow() != nrow()) {
+      report_error("Block must have the same number of rows as the "
+                   "ErrorExpanderMatrix.");
+    }
+    if (block.ncol() != ncol()) {
+      report_error("Block must have the same number of columns as the "
+                   "ErrorExpanderMatrix.");
+    }
+
+    size_t row_start = 0;
+    size_t col_start = 0;
+    for (const auto &b : blocks_) {
+      b->add_to_block(SubMatrix(
+          block, row_start, row_start + b->nrow() - 1,
+          col_start, col_start + b->ncol() - 1));
+      row_start += b->nrow();
+      col_start += b->ncol();
+    }
+  }
+
   SpdMatrix ErrorExpanderMatrix::inner() const {
     int dim = ncol();
     SpdMatrix ans(dim, 0.0);
