@@ -11,18 +11,28 @@ from .spikeslab import (
     plot_model_size,
 )
 
+from .mlogit_spike import mlogit_spike
+
 from .priors import (
     LogitZellnerPrior,
+    MultinomialLogitSpikeSlabPrior,
     RegressionSpikeSlabPrior,
     StudentSpikeSlabPrior,
 )
 
 __all__ = [
-    "dot",
-    "lm_spike",
+    "BigAssSpikeSlab",
     "LogitZellnerPrior",
+    "MultinomialLogitSpikeSlabPrior",
     "RegressionSpikeSlabPrior",
     "StudentSpikeSlabPrior",
-    "sparsify",
+    "coefficient_positive_probability",
+    "compute_inclusion_probabilities",
+    "dot",
+    "lm_spike",
+    "mlogit_spike",
+    "plot_inclusion_probs",
+    "plot_model_size",
     "set_glm_coefs",
+    "sparsify",
 ]
